package token

import "git.akyoto.dev/cli/q/src/build/keyword"

// Pre-allocate these byte buffers so we can re-use them
// instead of allocating a new buffer every time.
var (
	groupStartBytes = []byte{'('}
	groupEndBytes   = []byte{')'}
	blockStartBytes = []byte{'{'}
	blockEndBytes   = []byte{'}'}
	arrayStartBytes = []byte{'['}
	arrayEndBytes   = []byte{']'}
	separatorBytes  = []byte{','}
	newLineBytes    = []byte{'\n'}
)

// Tokenize turns the file contents into a list of tokens.
func Tokenize(buffer []byte) List {
	var (
		i      int
		tokens = make(List, 0, len(buffer)/2)
	)

	for i < len(buffer) {
		switch buffer[i] {
		// Whitespace
		case ' ', '\t':
		// Separator
		case ',':
			tokens = append(tokens, Token{Kind: Separator, Position: i, Bytes: separatorBytes})

		// Parentheses start
		case '(':
			tokens = append(tokens, Token{Kind: GroupStart, Position: i, Bytes: groupStartBytes})

		// Parentheses end
		case ')':
			tokens = append(tokens, Token{Kind: GroupEnd, Position: i, Bytes: groupEndBytes})

		// Block start
		case '{':
			tokens = append(tokens, Token{Kind: BlockStart, Position: i, Bytes: blockStartBytes})

		// Block end
		case '}':
			tokens = append(tokens, Token{Kind: BlockEnd, Position: i, Bytes: blockEndBytes})

		// Array start
		case '[':
			tokens = append(tokens, Token{Kind: ArrayStart, Position: i, Bytes: arrayStartBytes})

		// Array end
		case ']':
			tokens = append(tokens, Token{Kind: ArrayEnd, Position: i, Bytes: arrayEndBytes})

		// New line
		case '\n':
			tokens = append(tokens, Token{Kind: NewLine, Position: i, Bytes: newLineBytes})

		// Comment
		case '/':
			if i+1 >= len(buffer) || buffer[i+1] != '/' {
				position := i
				i++

				for i < len(buffer) && isOperator(buffer[i]) {
					i++
				}

				tokens = append(tokens, Token{Kind: Operator, Position: position, Bytes: buffer[position:i]})
				continue
			}

			position := i

			for i < len(buffer) && buffer[i] != '\n' {
				i++
			}

			tokens = append(tokens, Token{Kind: Comment, Position: position, Bytes: buffer[position:i]})
			continue

		// String
		case '"':
			start := i
			end := len(buffer)
			i++

			for i < len(buffer) {
				if buffer[i] == '"' {
					end = i + 1
					i++
					break
				}

				i++
			}

			tokens = append(tokens, Token{Kind: String, Position: start, Bytes: buffer[start:end]})
			continue

		default:
			// Identifier
			if isIdentifierStart(buffer[i]) {
				position := i
				i++

				for i < len(buffer) && isIdentifier(buffer[i]) {
					i++
				}

				identifier := buffer[position:i]
				keyword, isKeyword := keyword.Map[string(identifier)]

				if isKeyword {
					tokens = append(tokens, Token{Kind: Keyword, Position: position, Bytes: keyword})
				} else {
					tokens = append(tokens, Token{Kind: Identifier, Position: position, Bytes: identifier})
				}

				continue
			}

			// Number
			if isNumber(buffer[i]) {
				position := i
				i++

				for i < len(buffer) && isNumber(buffer[i]) {
					i++
				}

				tokens = append(tokens, Token{Kind: Number, Position: position, Bytes: buffer[position:i]})
				continue
			}

			// Operator
			if isOperator(buffer[i]) {
				position := i
				i++

				for i < len(buffer) && isOperator(buffer[i]) {
					i++
				}

				tokens = append(tokens, Token{Kind: Operator, Position: position, Bytes: buffer[position:i]})
				continue
			}

			// Invalid character
			tokens = append(tokens, Token{Kind: Invalid, Position: i, Bytes: buffer[i : i+1]})
		}

		i++
	}

	tokens = append(tokens, Token{Kind: EOF, Position: i, Bytes: nil})
	return tokens
}

func isIdentifier(c byte) bool {
	return isLetter(c) || isNumber(c) || c == '_'
}

func isIdentifierStart(c byte) bool {
	return isLetter(c) || c == '_'
}

func isLetter(c byte) bool {
	return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
}

func isNumber(c byte) bool {
	return (c >= '0' && c <= '9')
}

func isOperator(c byte) bool {
	switch c {
	case '=', ':', '.', '+', '-', '*', '/', '<', '>', '&', '|', '^', '%', '!':
		return true
	default:
		return false
	}
}
